from .admin_sidebar import *  # noqa
from .integration_models import *  # noqa
from .page_models import *  # noqa
from .snippet_models import *  # noqa
from .wagtailsettings_models import *  # noqa
from .cms_models import *  # noqa



from cjkcms.models.wagtailsettings_models import AdobeApiSettings
from cjkcms.settings import cms_settings
from django.contrib.staticfiles import finders
from django.template.loader import get_template
from django.test import TestCase
from wagtail.models import Site

# import TemplateDoesNotExist
from django.template import TemplateDoesNotExist

# Test the settings.py file and all templates and static files defined.


class SettingsTests(TestCase):
    def test_brand_logos(self):
        """Check if default brand logos are available."""

        logos = [
            cms_settings.CJKCMS_BRAND_LOGO_LONG,
            cms_settings.CJKCMS_BRAND_LOGO_SQUARE,
        ]

        for logo in logos:
            abs_path = finders.find(logo)
            # self.assertIsNone(abs_path, f"{logo} not found.")
            self.assertTrue(str(abs_path).endswith(logo), f"{logo} not found.")

    def _check_templates(self, templates, skip_empty=False):
        """Check if all declared template files specified in the `templates` list
        are accessible and execute without error. Used to check individual
        sections in settings.py"""

        for tpl in templates:
            if skip_empty and not tpl[0]:
                continue
            try:
                get_template(tpl[0])
            except TemplateDoesNotExist as e:
                self.fail(f"Template not found: {str(e)}")

    def test_cardblock_templates(self):
        templates = cms_settings.CJKCMS_FRONTEND_TEMPLATES_BLOCKS["cardblock"]
        self._check_templates(templates)

    def test_cardgridblock_templates(self):
        templates = cms_settings.CJKCMS_FRONTEND_TEMPLATES_BLOCKS["cardgridblock"]
        self._check_templates(templates)

    def test_pagelistblock_templates(self):
        templates = cms_settings.CJKCMS_FRONTEND_TEMPLATES_BLOCKS["pagelistblock"]
        self._check_templates(templates)

    def test_pagepreviewblock_templates(self):
        templates = cms_settings.CJKCMS_FRONTEND_TEMPLATES_BLOCKS["pagepreviewblock"]
        self._check_templates(templates)

    def test_frontend_template_pages_templates(self):
        templates = cms_settings.CJKCMS_FRONTEND_TEMPLATES_PAGES["*"]
        self._check_templates(templates, skip_empty=True)

    def test_AdobeApiKey(self):
        site = Site.objects.filter(is_default_site=True)[0]
        adobe_api_key = AdobeApiSettings.for_site(site=site)
        adobe_api_key.adobe_embed_id = "test_key"
        adobe_api_key.save()

        adobe_api_key = AdobeApiSettings.for_site(site=site)
        """Check if the Adobe API Key is set."""
        self.assertEqual(adobe_api_key.adobe_embed_id, "test_key")



from django.core.management.base import BaseCommand
from wagtail.models import Page, Site
from cjkcms.models import Navbar, LayoutSettings, NavbarOrderable


class Command(BaseCommand):
    """
    Create main-menu nav bar, and assign it in the Layout settings.
    Assigs a custom_id=main-menu to detect attempts to re-create.
    """

    help = "creates CjkCMS Navbar with Home entry pointing to /, "
    "and if --add_auth also login/logout entries"
    add_auth_items = False

    def add_arguments(self, parser):
        # Named (optional) arguments
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force overwrite if navbar with custom_id=main-menu already exists",
        )

        parser.add_argument(
            "--add_auth_items",
            action="store_true",
            help="Add entries for `/login`, `signup` and `[username]`",
        )

    def _setup_main_menu(self):
        n = Navbar.objects.filter(custom_id="main-menu")
        if len(n) > 0:
            print("main-menu navbar already exists, re-creating")
            n.delete()

        home_page = Page.objects.filter(path="00010001").first()
        if not home_page:
            print(
                "Homepage not found in page tree. "
                "Create at least the home page, then create main-menu"
            )
            return

        home_item = (("page_link", {"display_text": "Home", "page": home_page}),)
        auth_items = (
            (
                (
                    "external_link",
                    {
                        "display_text": "Login",
                        "link": "/login",
                        "visible_for": "non-auth-only",
                    },
                ),
                (
                    "external_link",
                    {
                        "display_text": "Register",
                        "link": "/signup",
                        "visible_for": "non-auth-only",
                    },
                ),
                (
                    "page_link",
                    {
                        "display_text": "{{ user.first_last_name }}",
                        "page": home_page,
                        "visible_for": "auth-only",
                    },
                ),
            )
            if self.add_auth_items
            else ()
        )

        n = Navbar(
            name="Main menu",
            custom_id="main-menu",
            menu_items=(home_item + auth_items),
        )
        n.save()

        # get default Wagtail site
        site = Site.objects.filter(is_default_site=True).first()
        layout_settings = LayoutSettings.for_site(site)

        # add NavbarOrderable to Layout
        no = NavbarOrderable(navbar_chooser=layout_settings, navbar=n)
        no.save()

    def handle(self, raise_error=False, *args, **options):
        # check if ok to overwrite (force == 1)
        force = options["force"]
        verbosity = options["verbosity"]
        self.add_auth_items = options["add_auth_items"]
        checks = [Navbar.objects.filter(custom_id="main-menu").count() > 0]
        if any(checks) and not force:
            # YOU SHOULD NEVER RUN THIS COMMAND WITHOUT PRIOR DB DUMP
            raise RuntimeError(
                "Navbar with custom_id=main-menu found. Aborting. use force=1 to overwrite"
            )

        self._setup_main_menu()
        if verbosity > 0:
            msg = "Main menu navbar successfully created."
            self.stdout.write(msg)



from wagtail.tests.utils import WagtailPageTests
from cjkcms.models import CjkcmsWebPage
from cjkcms.models.cms_models import WebPage
from wagtail.tests.utils.form_data import nested_form_data

from wagtail.models import Page


class WebPageTests(WagtailPageTests):
    """Tests CjkCms Web Page"""

    def test_cannot_create_cjkcms_web_page(self):
        # Get the HomePage
        # root_page = Page.objects.get(path='0001')
        # root page is useless, objects added here are not served under default site
        home_page = Page.objects.get(path="00010001")

        # Assert that a CjkcmsWebPage can NOT be made here, it's abstract
        self.assertCanNotCreateAt(
            home_page, CjkcmsWebPage, nested_form_data({"title": "About us"})
        )

    def test_can_create_web_page(self):
        # Get the HomePage
        body = [
            {
                "type": "embed_video",
                "value": {
                    "url": "https://youtu.be/AxLgTxBLsgI",
                },
                "id": "6ffa1b27-d884-4a9b-ba35-66140ff8a0ee",
            }
        ]
        home_page = Page.objects.get(path="00010001")

        web_page = WebPage(title="Test Web Home", body=body)

        home_page.add_child(instance=web_page)
        web_page.save_revision().publish()

        wp = WebPage.objects.get(slug="test-web-home")
        self.assertEqual(wp.title, "Test Web Home")

    # def test_site_root_is_polish(self):
    #     site = Site.objects.get()
    #     site.root_page.specific



"""
HTML blocks are simple blocks used to represent common HTML elements,
with additional styling and attributes.

HTML blocks should NOT contain more sub-blocks or sub-streamfields.
They must be safe to nest within more robust "content blocks" without
creating recursion.
"""
import logging

from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from wagtail.contrib.table_block.blocks import TableBlock as WagtailTableBlock
from wagtail import blocks
from wagtail.rich_text import get_text_for_indexing
from wagtail.documents.blocks import DocumentChooserBlock
from wagtail.embeds.blocks import EmbedBlock
from wagtail.images.blocks import ImageChooserBlock

from .base_blocks import (
    BaseBlock,
    BaseLinkBlock,
    ButtonMixin,
    CjkcmsAdvTrackingSettings,
    ClassifierTermChooserBlock,
    LinkStructValue,
)

logger = logging.getLogger("cjkcms")


class ButtonBlock(ButtonMixin, BaseLinkBlock):
    """
    A link styled as a button.
    """

    class Meta:
        template = "cjkcms/blocks/button_block.html"
        icon = "hand-pointer-o"
        label = _("Button Link")
        value_class = LinkStructValue


class DownloadBlock(ButtonMixin, BaseBlock):
    """
    Link to a file that can be downloaded.
    """

    automatic_download = blocks.BooleanBlock(
        required=False,
        label=_("Auto download"),
    )
    downloadable_file = DocumentChooserBlock(
        required=False,
        label=_("Document link"),
    )

    advsettings_class = CjkcmsAdvTrackingSettings

    class Meta:
        template = "cjkcms/blocks/download_block.html"
        icon = "download"
        label = _("Download")


class EmbedVideoBlock(BaseBlock):
    """
    Embedded media using stock wagtail functionality.
    """

    url = EmbedBlock(
        required=True,
        label=_("URL"),
        help_text=_("Link to a YouTube/Vimeo video, tweet, facebook post, etc."),
    )

    class Meta:
        template = "cjkcms/blocks/embed_video_block.html"
        icon = "media"
        label = _("Embed Media")


class H1Block(BaseBlock):
    """
    An <h1> heading.
    """

    text = blocks.CharBlock(
        max_length=255,
        label=_("Text"),
    )

    class Meta:
        template = "cjkcms/blocks/h1_block.html"
        icon = "header"
        label = _("Heading 1")


class H2Block(BaseBlock):
    """
    An <h2> heading.
    """

    text = blocks.CharBlock(
        max_length=255,
        label=_("Text"),
    )

    class Meta:
        template = "cjkcms/blocks/h2_block.html"
        icon = "header"
        label = _("Heading 2")


class H3Block(BaseBlock):
    """
    An <h3> heading.
    """

    text = blocks.CharBlock(
        max_length=255,
        label=_("Text"),
    )

    class Meta:
        template = "cjkcms/blocks/h3_block.html"
        icon = "header"
        label = _("Heading 3")


class TableBlock(BaseBlock):
    table = WagtailTableBlock()

    class Meta:
        template = "cjkcms/blocks/table_block.html"
        icon = "table"
        label = _("Table")


class ImageBlock(BaseBlock):
    """
    An <img>, by default styled responsively to fill its container.
    """

    image = ImageChooserBlock(
        label=_("Image"),
    )

    class Meta:
        template = "cjkcms/blocks/image_block.html"
        icon = "image"
        label = _("Image")
        label_format = "Image: {image}"


class ImageLinkBlock(BaseLinkBlock):
    """
    An <a> with an image inside it, instead of text.
    """

    image = ImageChooserBlock(
        label=_("Image"),
    )
    alt_text = blocks.CharBlock(
        max_length=255,
        required=True,
        help_text=_("Alternate text to show if the image doesn’t load"),
    )

    class Meta:
        template = "cjkcms/blocks/image_link_block.html"
        icon = "image"
        label = _("Image Link")
        value_class = LinkStructValue
        label_format = "Image link: {image}"


class PageListBlock(BaseBlock):
    """
    Renders a preview of selected pages.
    """

    indexed_by = blocks.PageChooserBlock(
        required=False,
        label=_("Parent page"),
        help_text=_(
            "Preview of pages that are children of the selected page (default to own parent). Uses ordering specified in the page’s LAYOUT tab."  # noqa
        ),
    )
    classified_by = ClassifierTermChooserBlock(
        required=False,
        label=_("Classified as"),
        help_text=_("Only show pages that are classified with this term."),
    )
    show_preview = blocks.BooleanBlock(
        required=False,
        default=False,
        label=_("Show body preview"),
    )
    show_dates = blocks.BooleanBlock(
        required=False,
        default=True,
        label=_("Show publish dates"),
    )

    show_authors = blocks.BooleanBlock(
        required=False,
        default=True,
        label=_("Show authors"),
    )

    show_covers = blocks.BooleanBlock(
        required=False,
        default=True,
        label=_("Show cover images"),
        help_text=_("Show cover images"),
    )

    show_navigation = blocks.BooleanBlock(
        required=False,
        default=False,
        label=_("Show navigation"),
        help_text=_("Show prev / next navigation buttons"),
    )

    num_posts = blocks.IntegerBlock(
        default=10,
        label=_("Number of pages to show"),
    )

    title = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_("Title"),
    )

    class Meta:
        template = "cjkcms/blocks/pagelist_block.html"
        icon = "list-ul"
        label = _("Latest Pages")

    def get_context(self, value, parent_context=None):

        context = super().get_context(value, parent_context=parent_context)

        if not parent_context:
            return context

        try:
            current_page = parent_context["page"]
        except KeyError:  # e.g. in search results
            return context

        if value["indexed_by"]:
            indexer = value["indexed_by"].specific
        else:
            indexer = current_page.get_parent().specific

        # try to use the CjkcmsPage `get_index_children()`,
        # but fall back to get_children if this is a non-CjkcmsPage
        if hasattr(indexer, "get_index_children"):
            pages = indexer.get_index_children().live()
            if value["classified_by"]:
                try:
                    pages = pages.filter(classifier_terms=value["classified_by"])
                except AttributeError:
                    # `pages` is not a queryset, or is not a queryset of CjkcmsPage.
                    logger.warning(
                        "Tried to filter by ClassifierTerm in PageListBlock, but <%s.%s ('%s')>.get_index_children()  # noqadid not return a queryset or is not a queryset of CjkcmsPage models.",  # noqa
                        indexer._meta.app_label,
                        indexer.__class__.__name__,
                        indexer.title,
                    )
        else:
            pages = indexer.get_children().live()

        pages_limited = pages[: value["num_posts"]]

        context["current_page"] = current_page

        # current_index = list(pages_limited.values_list("id", flat=True)).index(
        #     current_page.id
        # )
        # print(current_index)
        previous_page = None
        next_page = None
        current_index = -2

        page_count = len(list(pages_limited))
        for index, item in enumerate(pages_limited):
            if item.id == current_page.id:
                current_index = index

        for index, item in enumerate(pages_limited):
            print(index, current_index)
            if (index == current_index - 1) and (index >= 0):
                previous_page = item
            if (index == current_index + 1) and (index < page_count):
                next_page = item

        context["previous_page"] = previous_page
        context["next_page"] = next_page
        context["pages"] = pages_limited
        return context


class PagePreviewBlock(BaseBlock):
    """
    Renders a preview of a specific page.
    """

    page = blocks.PageChooserBlock(
        required=True,
        label=_("Page to preview"),
        help_text=_("Show a mini preview of the selected page."),
    )

    class Meta:
        template = "cjkcms/blocks/pagepreview_block.html"
        icon = "doc-empty-inverse"
        label = _("Page Preview")


class QuoteBlock(BaseBlock):
    """
    A <blockquote>.
    """

    text = blocks.TextBlock(
        required=True,
        rows=4,
        label=_("Quote Text"),
    )
    author = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_("Author"),
    )

    class Meta:
        template = "cjkcms/blocks/quote_block.html"
        icon = "openquote"
        label = _("Quote")


class RichTextBlock(blocks.RichTextBlock):
    class Meta:
        template = "cjkcms/blocks/rich_text_block.html"


class SearchableHTMLBlock(blocks.RawHTMLBlock):
    """RawHTMLBlock extended with basic search support"""

    def get_searchable_content(self, value):
        # Strip HTML tags to prevent search backend from indexing them
        source = force_str(value)
        return [get_text_for_indexing(source)]



"""
Layout blocks are essentially a wrapper around content.
e.g. rows, columns, hero units, etc.
"""

from cjkcms.blocks.base_blocks import BaseLayoutBlock, CjkcmsAdvColumnSettings
from cjkcms.settings import cms_settings
from django.utils.translation import gettext_lazy as _
from wagtail import blocks
from wagtail.images.blocks import ImageChooserBlock

# Level 1 layout blocks


class ColumnBlock(BaseLayoutBlock):
    """
    Renders content in a column.
    """

    column_size = blocks.ChoiceBlock(
        choices=cms_settings.CJKCMS_FRONTEND_COL_SIZE_CHOICES,
        default=cms_settings.CJKCMS_FRONTEND_COL_SIZE_DEFAULT,
        required=False,
        label=_("Column size"),
    )

    advsettings_class = CjkcmsAdvColumnSettings

    class Meta:
        template = "cjkcms/blocks/column_block.html"
        icon = "placeholder"
        label = _("Column")
        label_format = _("Column")


class GridBlock(BaseLayoutBlock):
    """
    Renders a row of columns.
    """

    fluid = blocks.BooleanBlock(
        required=False,
        label=_("Full width"),
    )

    class Meta:
        template = "cjkcms/blocks/grid_block.html"
        icon = "columns"
        label = _("Responsive Grid Row")
        label_format = _("Responsive Grid Row")

    def __init__(self, local_blocks=None, **kwargs):
        super().__init__(local_blocks=[("content", ColumnBlock(local_blocks))])


class CardGridBlock(BaseLayoutBlock):
    """
    Renders a row of cards.
    """

    fluid = blocks.BooleanBlock(
        required=False,
        label=_("Full width"),
    )

    default_card_template = blocks.ChoiceBlock(
        choices=cms_settings.CJKCMS_FRONTEND_TEMPLATES_BLOCKS["cardblock"],
        default=None,
        required=False,
        label=_("Card Template for this grid"),
        help_text=_(
            "Leave blank to use the default card template in this grid. "
            "Each card may override the grid-level default."
        ),
    )

    class Meta:
        template = "cjkcms/blocks/cardgrid_deck.html"
        icon = "th-large"
        label = _("Card Grid")
        label_format = _("Card Grid")


class HeroBlock(BaseLayoutBlock):
    """
    Wrapper with color and image background options.
    """

    fluid = blocks.BooleanBlock(
        required=False,
        default=True,
        label=_("Full width"),
    )
    is_parallax = blocks.BooleanBlock(
        required=False,
        label=_("Parallax Effect"),
        help_text=_(
            "Background images scroll slower than foreground images, creating an illusion of depth."
        ),  # noqa
    )
    background_image = ImageChooserBlock(required=False)
    tile_image = blocks.BooleanBlock(
        required=False,
        default=False,
        label=_("Tile background image"),
    )
    background_color = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_("Background color"),
        help_text=_("Hexadecimal, rgba, or CSS color notation (e.g. #ff0011)"),
    )
    foreground_color = blocks.CharBlock(
        required=False,
        max_length=255,
        label=_("Text color"),
        help_text=_("Hexadecimal, rgba, or CSS color notation (e.g. #ff0011)"),
    )

    class Meta:
        template = "cjkcms/blocks/hero_block.html"
        icon = "newspaper-o"
        label = "Hero Unit"



from django import template
from django.contrib.auth.models import Group

register = template.Library()


@register.filter(name="has_group")
def has_group(user, group_name):
    """Filter to check if current user belongs to specific group"""
    group = Group.objects.get(name=group_name)
    return group in user.groups.all()



from django import forms
from django.utils.translation import gettext_lazy as _


class SearchForm(forms.Form):
    s = forms.CharField(
        max_length=255,
        required=False,
        label=_("Search"),
    )
    t = forms.CharField(
        widget=forms.HiddenInput,
        max_length=255,
        required=False,
        label=_("Page type"),
    )



import json

from cjkcms.api.mailchimp import MailchimpApi
from django.db import models
from django.forms.widgets import Input
from django.template import Context, Template
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _
from wagtail import hooks
from wagtail.admin.panels import FieldPanel


class MailchimpSubscriberIntegrationWidget(Input):
    template_name = "cjkcms/formfields/mailchimp/subscriber_integration_widget.html"
    js_template_name = "cjkcms/formfields/mailchimp/subscriber_integration_js.html"

    def get_context(self, name, value, attrs):
        ctx = super(MailchimpSubscriberIntegrationWidget, self).get_context(
            name, value, attrs
        )

        json_value = self.get_json_value(value)
        list_library = self.build_list_library()
        ctx["widget"]["value"] = json.dumps(json_value)
        ctx["widget"]["extra_js"] = self.render_js(name, list_library, json_value)
        ctx["widget"][
            "selectable_mailchimp_lists"
        ] = self.get_selectable_mailchimp_lists(list_library)
        ctx["widget"]["stored_mailchimp_list"] = self.get_stored_mailchimp_list(
            json_value
        )

        return ctx

    def render_js(self, name, list_library, json_value):
        ctx = {
            "widget_name": name,
            "widget_js_name": name.replace("-", "_"),
            "list_library": list_library,
            "stored_mailchimp_list": self.get_stored_mailchimp_list(json_value),
            "stored_merge_fields": self.get_stored_merge_fields(json_value),
        }

        return render_to_string(self.js_template_name, ctx)

    def get_json_value(self, value):
        json_value = json.loads(value) if value else json.loads("{}")
        if "list_id" not in json_value:
            json_value["list_id"] = ""
        if "merge_fields" not in json_value:
            json_value["merge_fields"] = {}
        if "email_field" not in json_value:
            json_value["email_field"] = ""
        if "interest_categories" not in json_value:
            json_value["interest_categories"] = {}
        return json_value

    def get_stored_mailchimp_list(self, value):
        if "list_id" in value:
            return str(value["list_id"])

    def get_stored_merge_fields(self, value):
        if "merge_fields" in value:
            return json.dumps(value["merge_fields"])
        return json.dumps({})

    def get_selectable_mailchimp_lists(self, library):
        selectable_lists = [("", "Please select one of your Mailchimp Lists.")]
        selectable_lists.extend((k, v["name"]) for k, v in library.items())
        return selectable_lists

    def build_list_library(self):  # sourcery skip: merge-dict-assign
        mailchimp = MailchimpApi()
        list_library = {}
        if mailchimp.is_active:
            lists = mailchimp.get_lists()
            for mlist in lists["lists"]:
                list_library[mlist["id"]] = {
                    "name": mlist["name"],
                    "merge_fields": {},
                    "interest_categories": {},
                }

                list_library[mlist["id"]][
                    "merge_fields"
                ] = mailchimp.get_merge_fields_for_list(mlist["id"])["merge_fields"]
                list_library[mlist["id"]][
                    "interest_categories"
                ] = mailchimp.get_interest_categories_for_list(mlist["id"])[
                    "categories"
                ]

                for category in list_library[mlist["id"]]["interest_categories"]:
                    category[
                        "interests"
                    ] = mailchimp.get_interests_for_interest_category(
                        mlist["id"], category["id"]
                    )[
                        "interests"
                    ]

        return list_library


class MailchimpSubscriberIntegration(models.Model):
    class Meta:
        abstract = True

    subscriber_json_data = models.TextField(blank=True, verbose_name=_("List"))

    def integration_operation(self, instance, **kwargs):
        mailchimp = MailchimpApi()
        if mailchimp.is_active:
            submission_dict = kwargs["form_submission"].get_data()
            rendered_dictionary = self.render_dictionary(submission_dict)
            mailchimp.add_user_to_list(
                list_id=self.get_list_id(), data=rendered_dictionary
            )

    def get_data(self):
        return json.loads(self.subscriber_json_data)

    def get_merge_fields(self):
        if "merge_fields" in self.get_data():
            return self.get_data()["merge_fields"]
        return {}

    def get_list_id(self):
        if "list_id" in self.get_data():
            return self.get_data()["list_id"]

    def combine_interest_categories(self):
        # sourcery skip: dict-assign-update-to-union
        interest_dict = {}
        for category_id, value in self.get_data()["interest_categories"].items():
            interest_dict.update(value["interests"])

        return interest_dict

    def render_dictionary(self, form_submission):
        # sourcery skip: inline-immediately-returned-variable
        rendered_dictionary_template = json.dumps(
            {
                "members": [
                    {
                        "email_address": self.get_data()["email_field"],
                        "merge_fields": self.get_data()["merge_fields"],
                        "interests": self.combine_interest_categories(),
                        "status": "subscribed",
                    }
                ],
                "update_existing": True,
            }
        )

        rendered_dictionary = Template(rendered_dictionary_template).render(
            Context(form_submission)
        )
        return rendered_dictionary

    panels = [
        FieldPanel("subscriber_json_data", widget=MailchimpSubscriberIntegrationWidget)
    ]


@hooks.register("form_page_submit")
def run_mailchimp_subscriber_integrations(instance, **kwargs):
    if hasattr(instance, "integration_panels"):
        for panel in instance.integration_panels:
            for integration in getattr(instance, panel.relation_name).all():
                integration.integration_operation(instance, **kwargs)



"""
Base and abstract pages used in Cjk CMS.

Based on CODEREDCMS
"""

import logging
from datetime import datetime
from typing import Optional, TYPE_CHECKING

# import geocoder
from django import forms
from django.conf import settings
from django.core.paginator import Paginator, InvalidPage, EmptyPage, PageNotAnInteger
from django.db import models
from django.utils.translation import gettext_lazy as _

# from eventtools.models import BaseEvent, BaseOccurrence
# from icalendar import Event as ICalEvent
from modelcluster.fields import ParentalKey, ParentalManyToManyField
from modelcluster.contrib.taggit import ClusterTaggableManager
from taggit.models import TaggedItemBase
from wagtail.admin.panels import (
    FieldPanel,
    MultiFieldPanel,
    ObjectList,
    TabbedInterface,
)
from wagtail.fields import StreamField
from wagtail.models import PageBase, Page
from wagtail.coreutils import resolve_model_string
from wagtail.images import get_image_model_string
from wagtail.search import index
from wagtail.utils.decorators import cached_classmethod
from wagtailcache.cache import WagtailCacheMixin
from wagtailseo.models import SeoMixin, TwitterCard

from cjkcms import utils
from cjkcms.blocks import (
    CONTENT_STREAMBLOCKS,
    LAYOUT_STREAMBLOCKS,
)

from cjkcms.models.snippet_models import ClassifierTerm
from cjkcms.models.wagtailsettings_models import (
    LayoutSettings,
)

from cjkcms.settings import cms_settings
from cjkcms.widgets import ClassifierSelectWidget

if TYPE_CHECKING:
    from wagtail.images.models import AbstractImage

logger = logging.getLogger("cjkcms")


CJKCMS_PAGE_MODELS = []


def get_page_models():
    return CJKCMS_PAGE_MODELS


class CjkcmsPageMeta(PageBase):
    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        if "search_template" not in dct:
            cls.search_template = "cjkcms/pages/search_result.html"
        if not cls._meta.abstract:  # type: ignore
            CJKCMS_PAGE_MODELS.append(cls)


class CjkcmsTag(TaggedItemBase):
    class Meta:
        verbose_name = _("CMS Tag")

    content_object = ParentalKey("cjkcms.CjkcmsPage", related_name="tagged_items")


class CjkcmsPage(WagtailCacheMixin, SeoMixin, Page, metaclass=CjkcmsPageMeta):  # type: ignore
    """
    General use page with caching, templating, and SEO functionality.
    All pages should inherit from this.
    """

    class Meta:
        verbose_name = _("CjkCMS Page")

    # Do not allow this page type to be created in wagtail admin
    is_creatable = False

    # Templates
    # The page will render the following templates under certain conditions:
    #
    # template = ''
    # ajax_template = ''
    # search_template = ''

    ###############
    # Content fields
    ###############

    cover_image = models.ForeignKey(
        get_image_model_string(),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("Cover image"),
    )

    ###############
    # Index fields
    ###############

    # Subclasses can override this to enabled index features by default.
    index_show_subpages_default = False

    # Subclasses can override this to query on a specific
    # page model, rather than the default wagtail Page.
    index_query_pagemodel = "cjkcms.CjkcmsPage"

    # Subclasses can override these fields to enable custom
    # ordering based on specific subpage fields.
    index_order_by_default = ""
    index_order_by_choices = (
        ("", _("Default Ordering")),
        ("-first_published_at", _("Date first published, newest to oldest")),
        ("first_published_at", _("Date first published, oldest to newest")),
        ("-last_published_at", _("Date updated, newest to oldest")),
        ("last_published_at", _("Date updated, oldest to newest")),
        ("title", _("Title, alphabetical")),
        ("-title", _("Title, reverse alphabetical")),
    )
    index_show_subpages = models.BooleanField(
        default=index_show_subpages_default, verbose_name=_("Show list of child pages")
    )
    index_order_by_classifier = models.ForeignKey(
        "cjkcms.Classifier",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("Order child pages by classifier"),
        help_text=_(
            "Child pages will first be sorted following the order of this "
            "classifier’s terms (from Snippets > Classifiers)."
        ),
    )
    index_order_by = models.CharField(
        max_length=255,
        choices=index_order_by_choices,
        default=index_order_by_default,
        blank=True,
        verbose_name=_("Order child pages by"),
    )
    index_num_per_page = models.PositiveIntegerField(
        default=10,
        verbose_name=_("Number per page"),
    )
    index_classifiers = ParentalManyToManyField(
        "cjkcms.Classifier",
        blank=True,
        verbose_name=_("Filter child pages by"),
        help_text=_("Enable filtering child pages by these classifiers."),
    )

    ###############
    # Layout fields
    ###############

    custom_template = models.CharField(
        blank=True, max_length=255, choices=None, verbose_name=_("Template")
    )

    ###############
    # SEO overrides
    ###############

    seo_image_sources = [
        "og_image",
        "cover_image",
    ]

    ###############
    # Classify
    ###############

    classifier_terms = ParentalManyToManyField(
        "cjkcms.ClassifierTerm",
        blank=True,
        verbose_name=_("Classifiers"),
        help_text=_(
            "Categorize and group pages together with classifiers. "
            "Used to organize and filter pages across the site."
        ),  # noqa
    )
    tags = ClusterTaggableManager(
        through=CjkcmsTag,
        blank=True,
        verbose_name=_("Tags"),
        help_text=_("Used to organize pages across the site."),
    )

    ###############
    # Search
    ###############

    search_fields = Page.search_fields + [
        index.SearchField("seo_title", partial_match=True, boost=2),
        index.SearchField("search_description", boost=2),
        index.FilterField("index_show_subpages"),
        index.FilterField("index_order_by"),
        index.FilterField("custom_template"),
        index.FilterField("classifier_terms"),
    ]

    ###############
    # Panels
    ###############

    content_panels = Page.content_panels + [
        FieldPanel("cover_image"),
    ]

    body_content_panels = []

    bottom_content_panels = []

    classify_panels = [
        FieldPanel("classifier_terms", widget=ClassifierSelectWidget()),
        FieldPanel("tags"),
    ]

    layout_panels = [
        MultiFieldPanel([FieldPanel("custom_template")], heading=_("Visual Design")),
        MultiFieldPanel(
            [
                FieldPanel("index_show_subpages"),
                FieldPanel("index_num_per_page"),
                FieldPanel("index_order_by_classifier"),
                FieldPanel("index_order_by"),
                FieldPanel("index_classifiers", widget=forms.CheckboxSelectMultiple()),
            ],
            heading=_("Show Child Pages"),
        ),
    ]

    promote_panels = SeoMixin.seo_meta_panels + SeoMixin.seo_struct_panels

    integration_panels = []

    def __init__(self, *args, **kwargs):
        """
        Inject custom choices and defaults into the form fields
        to enable customization by subclasses.
        """
        super().__init__(*args, **kwargs)
        klassname = self.__class__.__name__.lower()
        template_choices = cms_settings.CJKCMS_FRONTEND_TEMPLATES_PAGES.get(
            "*", []
        ) + cms_settings.CJKCMS_FRONTEND_TEMPLATES_PAGES.get(klassname, [])

        self._meta.get_field("index_order_by").choices = self.index_order_by_choices  # type: ignore
        self._meta.get_field("custom_template").choices = template_choices  # type: ignore
        if not self.id:  # type: ignore
            self.index_order_by = self.index_order_by_default
            self.index_show_subpages = self.index_show_subpages_default

    @classmethod
    def get_panels(cls):  # sourcery skip: instance-method-first-arg-name
        panels = [
            ObjectList(
                cls.content_panels
                + cls.body_content_panels
                + cls.bottom_content_panels,
                heading=_("Content"),
            ),
            ObjectList(cls.classify_panels, heading=_("Classify")),
            ObjectList(cls.layout_panels, heading=_("Layout")),
            ObjectList(cls.promote_panels, heading=_("SEO"), classname="seo"),
            ObjectList(
                cls.settings_panels, heading=_("Settings"), classname="settings"
            ),
        ]

        if cls.integration_panels:
            panels.append(
                ObjectList(
                    cls.integration_panels,
                    heading="Integrations",
                    classname="integrations",
                )
            )
        return panels

    @cached_classmethod
    def get_edit_handler(cls):  # sourcery skip: instance-method-first-arg-name
        """
        Override to "lazy load" the panels overridden by subclasses.
        """
        panels = cls.get_panels()  # override e.g. by appending new panels
        edit_handler = TabbedInterface(panels)
        return edit_handler.bind_to_model(cls)

    @property
    def seo_logo(self) -> "Optional[AbstractImage]":
        # sourcery skip: assign-if-exp, or-if-exp-identity, use-named-expression
        """
        Gets logo for structured data using a fallback.
        Override method in SeoMixin.
        Gets the primary logo of the organization.
        """

        logo = super().seo_logo
        if logo:
            return logo
        layout_settings = LayoutSettings.for_site(self.get_site())
        if layout_settings.logo:
            return layout_settings.logo
        return None

    @property
    def seo_image(self) -> "Optional[AbstractImage]":
        """
        Override method in SeoMixin.
        Fallback to logo if opengraph image is not specified.
        """
        img = super().seo_image
        return self.seo_logo if img is None else img

    @property
    def seo_twitter_card_content(self) -> str:
        """
        Override of method in SeoMixin.
        Show a large twitter card if the page has an image set.
        """
        if self.seo_image:
            return TwitterCard.LARGE.value
        return self.seo_twitter_card.value

    def get_template(self, request, *args, **kwargs):
        """
        Override parent to serve different templates based on querystring.
        """
        return self.custom_template or super().get_template(request, args, kwargs)

    def get_index_children(self):
        """
        Returns query of subpages as defined by `index_` variables.
        """
        if self.index_query_pagemodel:
            querymodel = resolve_model_string(
                self.index_query_pagemodel, self._meta.app_label
            )
            query = querymodel.objects.child_of(self).live()  # type: ignore
        else:
            query = self.get_children().live()  # type: ignore

        # Determine query sorting order.
        order = []

        # To sort by term order of a specific classifier, annotate the child
        # pages with the `sort_order` of its ClassifierTerms.
        if self.index_order_by_classifier:
            terms = ClassifierTerm.objects.filter(
                classifier=self.index_order_by_classifier,
                # Reverse ManyToMany of `cjkcmspage.classifier_terms`.
                cjkcmspage=models.OuterRef("pk"),
            )
            query = query.annotate(
                term_sort_order=models.Subquery(terms.values("sort_order"))
            )
            order.append("term_sort_order")

        # Second, order by the specified model attribute.
        if self.index_order_by:
            order.append(self.index_order_by)

        # Order the query.
        if order:
            query = query.order_by(*order)

        return query

    def get_context(self, request, *args, **kwargs):
        """
        Add child pages and paginated child pages to context.
        """
        context = super().get_context(request)

        if self.index_show_subpages:
            # Get child pages
            all_children = self.get_index_children()
            # Filter by classifier terms if applicable
            if len(request.GET) > 0 and self.index_classifiers.exists():
                # Look up comma separated ClassifierTerm slugs i.e. `/?c=term1-slug,term2-slug`
                terms = []
                get_c = request.GET.get("c", None)
                if get_c:
                    terms = get_c.split(",")
                # Else look up individual querystrings i.e. `/?classifier-slug=term1-slug`
                else:
                    for classifier in self.index_classifiers.all().only("slug"):
                        get_term = request.GET.get(classifier.slug, None)
                        if get_term:
                            terms.append(get_term)
                if len(terms) > 0:
                    selected_terms = ClassifierTerm.objects.filter(slug__in=terms)
                    context["selected_terms"] = selected_terms
                    if len(selected_terms) > 0:
                        try:
                            for term in selected_terms:
                                all_children = all_children.filter(
                                    classifier_terms=term
                                )
                        except AttributeError:
                            logger.warning(
                                "Tried to filter by ClassifierTerm, but <%s.%s ('%s')>.get_index_children() did not return a queryset or is not a queryset of CjkcmsPage models.",  # noqa
                                self._meta.app_label,
                                self.__class__.__name__,
                                self.title,
                            )
            paginator = Paginator(all_children, self.index_num_per_page)
            pagenum = request.GET.get("p", 1)
            try:
                paged_children = paginator.page(pagenum)
            except (PageNotAnInteger, EmptyPage, InvalidPage) as e:  # noqa
                paged_children = paginator.page(1)

            context["index_paginated"] = paged_children
            context["index_children"] = all_children
        return context


###############################################################################
# Abstract pages providing pre-built common website functionality, suitable for subclassing.
# These are abstract so subclasses can override fields if desired.
###############################################################################


class CjkcmsWebPage(CjkcmsPage):
    """
    Provides a body and body-related functionality.
    This is abstract so that subclasses can override the body StreamField.
    """

    class Meta:
        verbose_name = _("CMS Web Page")
        abstract = True

    template = "cjkcms/pages/web_page.html"

    # Child pages should override based on what blocks they want in the body.
    # Default is LAYOUT_STREAMBLOCKS which is the fullest editor experience.

    body = StreamField(LAYOUT_STREAMBLOCKS, null=True, blank=True, use_json_field=True)

    # Search fields
    search_fields = CjkcmsPage.search_fields + [index.SearchField("body")]

    # Panels
    body_content_panels = [
        FieldPanel("body"),
    ]

    @property
    def body_preview(self):
        """
        A shortened version of the body without HTML tags.
        """
        return utils.get_richtext_preview(self.body)


class CjkcmsArticlePage(CjkcmsWebPage):
    """
    Article, suitable for news or blog content.
    """

    class Meta:
        verbose_name = _("CMS Article")
        abstract = True

    template = "cjkcms/pages/article_page.html"

    # Override body to provide simpler content
    body = StreamField(CONTENT_STREAMBLOCKS, null=True, blank=True, use_json_field=True)

    caption = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Caption"),
    )
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        editable=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Author"),
    )
    author_display = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Display author as"),
        help_text=_("Override how the author’s name displays on this article."),
    )
    date_display = models.DateField(
        null=True,
        blank=True,
        verbose_name=_("Display publish date"),
    )

    def seo_author(self) -> str:
        """
        Override of method in SeoMixin.
        Gets author name using a fallback.
        """
        if self.author_display:
            return self.author_display
        if self.author:
            return self.author.get_full_name()
        return self.owner.get_full_name() if self.owner else ""

    @property
    def seo_published_at(self) -> datetime:
        """
        Override of method in SeoMixin.
        Gets published date.
        """
        return self.date_display or self.first_published_at

    @property
    def seo_description(self) -> str:
        """
        Override of method in SeoMixin.
        Gets the description using a fallback.
        """
        return self.search_description or self.caption or self.body_preview or ""

    search_fields = CjkcmsWebPage.search_fields + [
        index.SearchField("caption", boost=2),
        index.FilterField("author"),
        index.FilterField("author_display"),
        index.FilterField("date_display"),
    ]

    content_panels = CjkcmsWebPage.content_panels + [
        FieldPanel("caption"),
        MultiFieldPanel(
            [
                FieldPanel("author"),
                FieldPanel("author_display"),
                FieldPanel("date_display"),
            ],
            _("Publication Info"),
        ),
    ]


class CjkcmsArticleIndexPage(CjkcmsWebPage):
    """
    Shows a list of article sub-pages.
    """

    class Meta:
        verbose_name = _("CMS Article Index Page")
        abstract = True

    template = "cjkcms/pages/article_index_page.html"

    index_show_subpages_default = True

    index_order_by_default = "-date_display"
    index_order_by_choices = (
        ("-date_display", "Display publish date, newest first"),
    ) + CjkcmsWebPage.index_order_by_choices

    show_images = models.BooleanField(
        default=True,
        verbose_name=_("Show images"),
    )
    show_captions = models.BooleanField(
        default=True,
    )
    show_meta = models.BooleanField(
        default=True,
        verbose_name=_("Show author and date info"),
    )
    show_preview_text = models.BooleanField(
        default=True,
        verbose_name=_("Show preview text"),
    )

    layout_panels = CjkcmsWebPage.layout_panels + [
        MultiFieldPanel(
            [
                FieldPanel("show_images"),
                FieldPanel("show_captions"),
                FieldPanel("show_meta"),
                FieldPanel("show_preview_text"),
            ],
            heading=_("Child page display"),
        ),
    ]


# !GK events removed from CjkCMS (see Codered CRX if you need them)
# !GK forms removed from CjkCMS (see Codered CRX if you need them)
# !GK location models removed from CjkCMS (see Codered CRX if you need them)
